/* 
 * Listing 3 - demo_client.c
 */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <time.h>
#include <rpc/rpc.h>

#include "ips.h"

/* Prototypes
 */
char *get_file      (void);
int   decode        (CLIENT *cl);
int   encode        (CLIENT *cl);
int   get_operators (CLIENT *cl);
int   get_loaders   (CLIENT *cl);
int   get_savers    (CLIENT *cl);
void  image_release (CLIENT *cl);
void  upcase        (char *text);
void  help_text     (void);

static u_long         image_id = -1;
static struct timeval timeout  = { 10000, 0 };

int main (int argc, char **argv)
{
  CLIENT *cl;
  Packet *request, *reply;
  int     id, done = 0;
  char    cmd[256];
  char    command[256];

  request = malloc (sizeof(Packet));
  if (argc < 2)
  {
    printf ("Usage: %s server\n", argv[0]);
    return (0);
  }

  cl = clnt_create (argv[1], IPSPROG, IPSVERS, "tcp");
  if (!cl)
  {
    clnt_pcreateerror (argv[1]);
    return (1);
  }

  clnt_control (cl, CLSET_TIMEOUT, (char *) &timeout);

  request->op = IPS_INIT;
  reply = ips_1 (request, cl);

  if (reply)
     printf ("\n%s\n\n", reply->Packet_u.hello_string);

  while (!done)
  {
    printf ("(? for help)>>");
    fflush (stdout);
    gets (cmd);

    if (!strcmp(cmd,"help") ||
        !strcmp(cmd,"?"))
      help_text();

    else if (!strcmp(cmd,"send"))
      decode (cl);

    else if (!strcmp(cmd,"get"))
      encode (cl);

    else if (!strcmp(cmd,"process"))
      process (cl);

    else if (!strcmp(cmd,"operators"))
      get_operators (cl);

    else if (!strcmp(cmd,"loaders"))
      get_loaders (cl);

    else if (!strcmp(cmd,"savers"))
      get_savers (cl);

    else if (!strcmp(cmd,"quit"))
    {
      image_release (cl);
      done = 1;
    }
    printf ("\n");
  }
  return (0);
}


int encode (CLIENT *cl)
{
  Packet *reply, *request;
  char    buf[256];
  char    filename[256];
  char    format[40];

  request = malloc (sizeof(Packet));
  printf ("\nIPS Send Image:\n\n");

  printf ("Enter image filename (abort to quit): ");
  fflush (stdout);
  gets (filename);
  if (!strcmp (filename, "abort"))
    return (1);

  printf ("Enter format (abort to quit): ");
  fflush (stdout);
  gets (format);
  if (!strcmp (format, "abort"))
    return (1);

  upcase (format);

  request->op = IPS_REQUEST_IMAGE;
  request->Packet_u.send_ops.id        = image_id;
  request->Packet_u.send_ops.format    = format;
  request->Packet_u.send_ops.argc      = 0;
  request->Packet_u.send_ops.argv.name = "test";
  request->Packet_u.send_ops.argv.next = NULL;

  reply = ips_1 (request, cl);
  if (!reply)
  {
    puts ("Reply error or timeout.");
    exit (1);
  }

  if (reply->op == IPS_SEND_IMAGE)
  {
    FILE *fp = fopen (filename, "w");
    fwrite (reply->Packet_u.img.data.data_val, 1,
	    reply->Packet_u.img.data.data_len, fp);
    fclose (fp);
    printf ("\"%s\" Received.\n", filename);
  }
  else if (reply->op == IPS_IMAGE_NOT_AVAIL)
    printf ("IPS_IMAGE_NOT_AVAIL recevied.\n");

  else if (reply->op == IPS_UNSUPPORTED_FORMAT)
    printf ("IPS_UNSUPPORTED_FORMAT received\n");

  else
  {
    printf ("Error in encoding.\n");
    return (1);
  }

  free (request);
  if (reply)
    xdr_free (xdr_Packet, reply);

  return (0);
}


int decode (CLIENT *cl)
{
  Packet *reply, *request;
  FILE   *fp;
  long    size;
  char    buf [256];
  char    buf2[256];
  char   *filename;
  u_char *raw_img;

  request = malloc (sizeof (Packet));

  if ((filename = get_file()) == NULL)
  {
    printf ("Error or aborted.\n");
    return (1);
  }

  printf ("Please enter image format: ");
  fflush (stdout);
  scanf ("%s", buf2);

  upcase (buf2);
  image_release (cl);

  sprintf (buf, "%s", filename);

  fp = fopen (buf, "r");
  fseek (fp, 0, 2);
  size = ftell (fp);

  fseek (fp, 0, 0);

  raw_img = malloc (size);
  fread (raw_img, 1, size, fp);
  fclose (fp);

  request->op = IPS_DECODE;
  request->Packet_u.raw_img.filename      = filename;
  request->Packet_u.raw_img.format        = strdup (buf2);
  request->Packet_u.raw_img.data.data_len = size;
  request->Packet_u.raw_img.data.data_val = (char*) raw_img;

  reply = ips_1 (request, cl);
  if (!reply)
  {
    puts ("Timeout.");
    exit (1);
  }

  if (reply->op == IPS_OK_ID)
     image_id = reply->Packet_u.id;
  else
  {
    printf ("Error in decoding.\n\n");
    return (1);
  }

  free (raw_img);
  free (request);
  xdr_free (xdr_Packet, reply);
  return (0);
}


int process (CLIENT *cl)
{
  Packet *reply = NULL, *request;
  int     i, retval, nitems;
  list   *nlp, *cp;
  char    buf[256];
  char   *items[10];

  if (image_id == -1)
  {
    printf ("No image sent to server.\n");
    return;
  }

  /*
   * For this article, nitems will be 0, but
   * this allows you to send arguments to the
   * operator program.  For example, if you
   * were executing a SCALE program, you would
   * have to send the new width and height.
   */
  nitems  = 0;
  request = malloc (sizeof (Packet));

  printf ("Please enter processing command: ");
  fflush (stdout);
  scanf ("%s", buf);

  upcase (buf);

  request->op = IPS_PROCESS;
  request->Packet_u.proc.id = image_id;
  request->Packet_u.proc.command = strdup (buf);
  request->Packet_u.proc.argc = nitems;

  if (nitems == 0)
  {
    request->Packet_u.proc.argv.name = "\0";
    request->Packet_u.proc.argv.next = 0;
  }
  else
  {
    nlp = &request->Packet_u.proc.argv;
    for (i = 0; i < nitems; i++)
    {
      nlp->name = strdup (items[i]);
      cp = nlp;
      nlp->next = malloc (sizeof(list));
      nlp = nlp->next;
    }
    cp->next = nlp = NULL;
  }

  reply = ips_1 (request, cl);

  free (request);

  if (!reply)
  {
    puts ("Encode timeout.");
    return (-1);
  }

  if (reply->op == IPS_PROCESS_OK)
    retval = 0;
  else
  {
    puts ("There was an error processing.\n");
    retval = -1;
  }

  xdr_free (xdr_Packet, reply);
  return (retval);
}


int get_loaders (CLIENT *cl)
{
  Packet *reply, *request;
  list   *nlp;

  request = malloc (sizeof (Packet));
  request->op = IPS_REQUEST_LOADERS;
  reply = ips_1 (request, cl);

  switch (reply->op)
  {
    case IPS_SEND_LOADERS:
	 printf ("\nIPS Load Formats:\n\n");
         for (nlp = &reply->Packet_u.operators; nlp; nlp = nlp->next)
            printf (" %s\n", nlp->name);
	 printf ("\n");
	 break;
     case ERROR:
	 printf ("get_loaders failed.\n");
         return (1);
  }

  free (request);
  xdr_free (xdr_Packet, reply);
  return (0);
}

int get_savers (CLIENT *cl)
{
  Packet *reply, *request;
  list   *nlp;

  request = malloc (sizeof (Packet));
  request->op = IPS_REQUEST_SAVERS;
  reply = ips_1 (request, cl);

  switch (reply->op)
  {
    case IPS_SEND_SAVERS:
	 printf ("\nIPS Save Formats:\n\n");
         for (nlp = &reply->Packet_u.operators; nlp; nlp = nlp->next)
            printf (" %s\n", nlp->name);
	 printf ("\n");
	 break;
    case ERROR:
	 printf ("get_savers failed.\n");
         return (1);
  }

  free (request);
  xdr_free (xdr_Packet, reply);
  return (0);
}

int get_operators (CLIENT *cl)
{
  Packet *reply, *request;
  list   *nlp;

  request = malloc (sizeof (Packet));

  request->op = IPS_REQUEST_OPS;
  reply = ips_1 (request, cl);

  switch (reply->op)
  {
    case IPS_SEND_OPS:
	 printf ("\nIPS Processing Tools:\n\n");
         for (nlp = &reply->Packet_u.operators; nlp; nlp = nlp->next)
	   printf (" %s\n", nlp->name);
	 printf ("\n");
	 break;
    case ERROR:
	 printf ("get_operators failed.\n");
         return (1);
  }

  free (request);
  xdr_free (xdr_Packet, reply);
  return (0);
}


char *get_file (void)
{
  int  status;
  char filename[1024], command[1024];

  printf ("\nIPS Receive Image:\n\n");
  printf ("Type 'abort' to cancel.  Filename: ");
  fflush (stdout);
  gets (filename);

  if (!strcmp (filename, "abort"))
     return (NULL);

  if (!access (filename, R_OK))
     return strdup (filename);

  return (NULL);
}


void image_release (CLIENT *cl)
{
  Packet *request, *reply;

  request = malloc (sizeof (Packet));
  if (image_id != -1)
  {
    request->op = IPS_IMAGE_RELEASE;
    request->Packet_u.id = image_id;
    reply = ips_1 (request, cl);
    if (reply->op == IPS_RELEASE_OK)
       printf ("Release of %d ok.\n", image_id);
  }
  free (request);
}  

void upcase (char *text)
{
  char *p = text;

  while (p && *p)
  {
    *p = toupper (*p);
    p++;
  }
}

void help_text (void)
{
  printf ("\n");
  printf ("  Cmd           Description\n");
  printf ("--------   -----------------------\n");
  printf (" send      send an image to the IPS\n");
  printf (" get       receive an image from IPS\n");
  printf ("           in a certain format\n");
  printf (" process   process an image\n");
  printf (" loaders   list all available loaders\n");
  printf (" savers    list all available savers\n");
  printf (" operators list all avail. operators\n");
  printf (" quit      exit this client\n");
  printf ("\n");
}
