/*
 * RE_SID: @(%)/tmp_mnt/vol/dosnfs/shades_SCCS/unix/pcnfsd/v2/src/SCCS/
 * s.pcnfsd_print.c 1.12 93/01/29 17:53:03 SMI
 */

/*
**=====================================================================
** Copyright (c) 1986-1993 by Sun Microsystems, Inc.
**	@(#)pcnfsd_print.c	1.12	1/29/93
**=====================================================================
*/

/*
**=====================================================================
**             I N C L U D E   F I L E   S E C T I O N                *
**                                                                    *
** If your port requires different include files, add a suitable      *
** #define in the customization section, and make the inclusion or    *
** exclusion of the files conditional on this.                        *
**=====================================================================
*/
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <rpc/rpc.h>
#include <netdb.h>
#include <pwd.h>

#include "common.h"
#include "pcnfsd.h"
#include "extern.h"

/* 
 * The following defintions give the maximum time allowed for
 * an external command to run (in seconds)
 */
#define MAXTIME_FOR_PRINT	10
#define MAXTIME_FOR_QUEUE	10
#define MAXTIME_FOR_CANCEL	10
#define MAXTIME_FOR_STATUS	10

#define QMAX 50

/* 
 * The following is derived from ucb/lpd/displayq.c
 */
#define SIZECOL 62
#define FILECOL 24

struct stat statbuf;
char   pathname    [MAXPATHLEN];
char   new_pathname[MAXPATHLEN];
char   sp_name     [MAXPATHLEN] = SPOOLDIR;
char   tempstr[256];
char   delims[] = " \t\r\n:()";

pr_list  printers = NULL;
pr_queue queue    = NULL;

#ifdef ACCEPT_BSD_PR_NAMES
pr_list alt_printers = NULL;
#endif

/*
**=====================================================================
**                      C O D E   S E C T I O N                       *
**=====================================================================
**
**=====================================================================
**                    C A V E A T   L E C T O R  ! ! ! !              *
**                                                                    *
** The following code relies upon the definition or non-definition of *
** a set of manifest constants which control whether BSD or SVR4      *
** rules should be adopted. It's not as simple as it might be,        *
** because certain OS's may support a mixture of facilities (or in    *
** some cases a mechanism may be broken).                             *
** This mechanism also allows for the addition of new styles.         *
**                                                                    *
** Within each line, only one symbol should be defined.               *
**                                                                    *
**  SVR4_STYLE_PRINT     BSD_STYLE_PRINT                              *
**  SVR4_STYLE_PR_LIST   BSD_STYLE_PR_LIST                            *
**  SVR4_STYLE_QUEUE     BSD_STYLE_QUEUE                              *
**  SVR4_STYLE_CANCEL    BSD_STYLE_CANCEL                             *
**  SVR4_STYLE_STATUS    BSD_STYLE_STATUS                             *
**  SVR4_STYLE_MONITOR   BSD_STYLE_MONITOR                            *
**                                                                    *
** The actual per-OS definitions are given in common.h, and are       *
** based on the definition of OSVER_xxx in the Makefile             *
**=====================================================================
*/

/* 
 * valid_pr
 * true if a printer name is valid
 */
int valid_pr (char *pr)
{
  char   *p;
  pr_list curr;

  mon_printers();

  if (printers == NULL)
     build_pr_list();

  if (printers == NULL)
     return (0);               /* can't tell - assume it's bad */

  p = map_printer_name (pr);
  if (p == NULL)
     return (1);               /* must be ok if maps to NULL! */

  curr = printers;
  while (curr)
  {
    if (!strcmp (p, curr->pn))
       return (1);
    curr = curr->pr_next;
  }

#ifdef ACCEPT_BSD_PR_NAMES
  if (alt_printers == NULL)
     return (0);               /* No alt list  */

  curr = alt_printers;
  while (curr)
  {
    if (!strcmp(p, curr->pn))
       return (1);
    curr = curr->pr_next;
  }
#endif
  return (0);
}


/* 
 * pr_init
 *
 * create a spool directory for a client if necessary and return the info
 */
pirstat pr_init (char *sys, char *pr, char **sp)
{
  int dir_mode = 0777;
  int rc;

  *sp = &pathname[0];
  pathname[0] = '\0';

  if (suspicious (sys) || suspicious (pr))
     return (PI_RES_FAIL);

  /* get pathname of current directory and return to client
   */
  sprintf (pathname, "%s/%s", sp_name, sys);
  mkdir (sp_name, dir_mode);         /* ignore the return code */
  chmod (sp_name, dir_mode);
  rc = mkdir (pathname, dir_mode);   /* DON'T ignore this return code */

  if ((rc < 0 && errno != EEXIST) ||
      (chmod (pathname, dir_mode) != 0) ||
      (stat (pathname, &statbuf) != 0) ||
      !(statbuf.st_mode & S_IFDIR))
  {
    sprintf (tempstr, "rpc.pcnfsd: unable to set up spool directory %s\n",
             pathname);
    msg_out (tempstr);
    pathname[0] = '\0';		/* null to tell client bad vibes */
    return (PI_RES_FAIL);
  }
  if (!valid_pr (pr))
  {
    pathname[0] = '\0';		/* null to tell client bad vibes */
    return (PI_RES_NO_SUCH_PRINTER);
  }
  return (PI_RES_OK);
}


/* 
 * pr_start2
 * start a print job
 */
psrstat pr_start2 (char *system, char *pr,   char  *user,
                   char *fname,  char *opts, char **id)
{
  static char req_id[256];
  char   snum[20];
  char   cmdbuf[256];
  char   resbuf[256];
  char  *xcmd;
  FILE  *fd;
  int    i, failed = 0;

  *id = &req_id[0];
  req_id[0] = '\0';

  if (suspicious(system) || suspicious(pr) ||
      suspicious(user)   || suspicious(fname))
    return (PS_RES_FAIL);

  sprintf (pathname, "%s/%s/%s", sp_name, system, fname);

  if (stat (pathname, &statbuf))
  {
    /* We can't stat the file. Let's try appending '.spl' and
     * see if it's already in progress.
     */
    strcat (pathname, ".spl");
    if (stat (pathname, &statbuf))
    {
      /* It really doesn't exist.
       */
      return (PS_RES_NO_FILE);
    }
    /* It is already on the way.
     */
    return (PS_RES_ALREADY);
  }

  if (statbuf.st_size == 0)
  {
    /* Null file - don't print it, just kill it.
     */
    unlink (pathname);
    return (PS_RES_NULL);
  }

  /*
   * The file is real, has some data, and is not already going out.
   * We rename it by appending '.spl' and exec "lpr" to do the
   * actual work.
   */
  strcpy (new_pathname, pathname);
  strcat (new_pathname, ".spl");

  /* See if the new filename exists so as not to overwrite it.
   */
  if (!stat (new_pathname, &statbuf))
  {
    strcpy (new_pathname, pathname);     /* rebuild a new name */
    sprintf (snum, "%d", rand());        /* get some number */
    strncat (new_pathname, snum, 3);
    strcat (new_pathname, ".spl");       /* new spool file */
  }
  if (rename (pathname, new_pathname))
  {
    /* Should never happen.
     */
    sprintf (tempstr, "rpc.pcnfsd: spool file rename (%s->%s) failed.\n",
             pathname, new_pathname);
    msg_out (tempstr);
    return (PS_RES_FAIL);
  }

  if (*opts == 'd')
  {
    /* This is a Diablo print stream. Apply the ps630
     * filter with the appropriate arguments.
     */
    run_ps630 (new_pathname, opts);
  }

  /* Try to match to an aliased printer
   */
  xcmd = expand_alias (pr, new_pathname, user, system);

  /* Use the copy option so we can remove the orignal spooled
   * nfs file from the spool directory.
   */
  if (!xcmd)
  {
    sprintf (cmdbuf, "/usr/bin/lpr -P%s -C%s -J%s %s",
	     pr, system, user, new_pathname);
    xcmd = cmdbuf;
  }
  if ((fd = su_popen (user, xcmd, MAXTIME_FOR_PRINT)) == NULL)
  {
    msg_out ("rpc.pcnfsd: su_popen failed");
    return (PS_RES_FAIL);
  }
  req_id[0] = '\0';		/* asume failure */

  while (fgets(resbuf, 255, fd))
  {
    i = strlen (resbuf);
    if (i)
       resbuf[i-1] = '\0';     /* trim NL */

    if (!strncmp (resbuf, "request id is ", 14))
    {
      /* New - just the first word is needed */
      strcpy (req_id, strtok (&resbuf[14], delims));
    }
    else if (strembedded ("disabled", resbuf))
      failed = 1;
  }
  if (su_pclose (fd) == 255)
     msg_out ("rpc.pcnfsd: su_pclose alert");

  unlink (new_pathname);
  return ((failed | interrupted) ? PS_RES_FAIL : PS_RES_OK);
}


/* 
 * Decide whether to blow away the printers list based on
 * a change to a specified file or directory.
 */
void mon_printers (void)
{
  static struct stat buf;
  static time_t old_mtime = 0;
  char  *mon_path = "/etc/printcap";

  if (stat (mon_path, &buf) == -1)
     return;

  if (old_mtime)
  {
    if (old_mtime != buf.st_mtime && printers)
    {
      free_pr_list_item (printers);
      printers = NULL;
    }
  }
  old_mtime = buf.st_mtime;
}


int build_pr_list (void)
{
  pr_list last = NULL;
  pr_list curr = NULL;
  char    buff[256];
  char   *cp;
  FILE   *p;

  p = popen ("/usr/sbin/lpc status", "r");
  if (p == NULL)
  {
    printers = list_virtual_printers();
    return (1);
  }

  while (fgets (buff, 255, p))
  {
  /*
   * This logic presumes that it's sufficient to find lines with
   * a non-space first character. We should actually check for
   * lines that begin "name:" - later
   */
    if (isspace (buff[0]))
       continue;

    cp = strtok (buff, delims);
    if (!cp)
       continue;
    curr = (struct pr_list_item*) grab (sizeof(struct pr_list_item));

    curr->pn      = my_strdup (cp);
    curr->device  = NULL;
    curr->remhost = NULL;
    curr->cm      = my_strdup ("-");
    curr->pr_next = NULL;
    curr->device  = my_strdup (cp);
    curr->remhost = my_strdup ("");
    if (last == NULL)
         printers = curr;
    else last->pr_next = curr;
    last = curr;

  }
  pclose (p);

  /*
   * Now add on the virtual printers, if any
   */
  if (last == NULL)
       printers = list_virtual_printers();
  else last->pr_next = list_virtual_printers();
  return (1);
}

/* 
 * free_pr_list_item
 *
 * free a list of printers recursively, item by item
 */
void free_pr_list_item (pr_list curr)
{
  if (curr->pn)
     free (curr->pn);
  if (curr->device)
     free (curr->device);
  if (curr->remhost)
     free (curr->remhost);
  if (curr->cm)
     free (curr->cm);
  if (curr->pr_next)
     free_pr_list_item (curr->pr_next);  /* recurse */
  free (curr);
}


/* 
 * BSD style queue listing
 */
pirstat build_pr_queue (pn, user, just_mine, p_qlen, p_qshown)
        printername pn;
        username    user;
        int         just_mine;
        int        *p_qlen;
        int        *p_qshown;
{
  pr_queue last = NULL;
  pr_queue curr = NULL;
  char     buff[256];
  FILE    *p;
  char    *cp, *rank, *owner;
  char    *job, *files, *totsize;
  int      i;

  if (queue)
  {
    free_pr_queue_item (queue);
    queue = NULL;
  }
  *p_qlen   = 0;
  *p_qshown = 0;
  pn = map_printer_name (pn);
  if (pn == NULL || suspicious (pn) || !valid_pr (pn))
     return (PI_RES_NO_SUCH_PRINTER);

  sprintf (buff, "/usr/ucb/lpq -P%s", pn);

  p = su_popen (user, buff, MAXTIME_FOR_QUEUE);
  if (p == NULL)
  {
    msg_out ("rpc.pcnfsd: unable to popen() lpq");
    return (PI_RES_FAIL);
  }

  while (fgets (buff, 255, p))
  {
    i = strlen (buff) - 1;
    buff[i] = '\0';		/* zap trailing NL */
    if (i < SIZECOL)
       continue;

    if (!strnicmp (buff, "rank", 4))
       continue;

    if (!strnicmp (buff, "warning", 7))
       continue;

    totsize = &buff[SIZECOL - 1];
    files   = &buff[FILECOL - 1];
    cp = totsize;
    cp--;
    while (cp > files && isspace (*cp))
      *cp-- = '\0';

    buff[FILECOL-2] = '\0';

    cp = strtok (buff, delims);
    if (!cp)
       continue;

    rank = cp;
    cp   = strtok (NULL, delims);
    if (!cp)
      continue;

    owner = cp;
    cp = strtok (NULL, delims);
    if (!cp)
       continue;

    job = cp;
    *p_qlen += 1;

    if (*p_qshown > QMAX)
      continue;

    if (just_mine && stricmp (owner, user))
      continue;

    *p_qshown += 1;

    curr = (struct pr_queue_item*) grab (sizeof(struct pr_queue_item));

    curr->position = atoi (rank);	/* active -> 0 */
    curr->id       = my_strdup (job);
    curr->size     = my_strdup (totsize);
    curr->status   = my_strdup (rank);
    curr->system   = my_strdup ("");
    curr->user     = my_strdup (owner);
    curr->file     = my_strdup (files);
    curr->cm       = my_strdup ("-");
    curr->pr_next  = NULL;

    if (last == NULL)
         queue = curr;
    else last->pr_next = curr;
    last = curr;
  }
  su_pclose (p);
  return (PI_RES_OK);
}


/* 
 * free_pr_queue_item
 *
 * recursively free a pr_queue, item by item
 */
void free_pr_queue_item (pr_queue curr)
{
  if (curr->id)
     free (curr->id);
  if (curr->size)
     free (curr->size);
  if (curr->status)
     free (curr->status);
  if (curr->system)
     free (curr->system);
  if (curr->user)
     free (curr->user);
  if (curr->file)
     free (curr->file);
  if (curr->cm)
     free (curr->cm);
  if (curr->pr_next)
     free_pr_queue_item (curr->pr_next);   /* recurse */
  free (curr);
}


pirstat get_pr_status (pn, avail, printing, qlen, needs_operator, status)
        printername pn;
        bool_t     *avail;
        bool_t     *printing;
        int        *qlen;
        bool_t     *needs_operator;
        char       *status;
{
  char    cmd[128];
  char    buff[256];
  char    buff2[256];
  char    pname[64];
  FILE   *p;
  char   *cp, *cp1, *cp2;
  int     n;
  pirstat stat = PI_RES_NO_SUCH_PRINTER;

  *avail    = FALSE;         /* assume the worst */
  *printing = FALSE;
  *needs_operator = FALSE;
  *qlen   = 0;
  *status = '\0';

  pn = map_printer_name (pn);
  if (pn == NULL || suspicious(pn) || !valid_pr(pn))
     return (PI_RES_NO_SUCH_PRINTER);

  n = sprintf (pname, "%s:", pn);
  sprintf (cmd, "/usr/sbin/lpc status %s", pn);
  p = popen (cmd, "r");
  if (p == NULL)
  {
    msg_out ("rpc.pcnfsd: unable to popen() lpc status");
    return (PI_RES_FAIL);
  }

  while (fgets (buff, 255, p))
  {
    if (strncmp (buff, pname, n))
      continue;

    /* We have a match. The only failure now is PI_RES_FAIL if
     * lpstat output cannot be decoded
     */
    stat = PI_RES_FAIL;

    /*
     * The next four lines are usually if the form
     *
     *     queuing is [enabled|disabled]
     *     printing is [enabled|disabled]
     *     [no entries | N entr[y|ies] in spool area]
     *     <status message, may include the word "attention">
     */
    while (fgets (buff, 255,p) && isspace (buff[0]))
    {
      cp = buff;
      while (isspace (*cp))
	cp++;
      if (*cp == '\0')
	break;
      cp1 = cp;
      cp2 = buff2;
      while (*cp1 && *cp1 != '\n')
      {
	*cp2++ =
	  (isupper (*cp1) ? tolower (*cp1) : *cp1);
	cp1++;
      }
      *cp1 = '\0';
      *cp2 = '\0';

      /*
       * Now buff2 has a lower-cased copy and cp points at the original;
       * both are null terminated without any newline
       */
      if (!strncmp (buff2, "queuing", 7))
      {
	*avail = (strstr (buff2, "enabled") != NULL);
	continue;
      }
      if (!strncmp (buff2, "printing", 8))
      {
	*printing = (strstr (buff2, "enabled") != NULL);
	continue;
      }
      if (isdigit (buff2[0]) && (strstr (buff2, "entr") != NULL))
      {
        *qlen = atoi (buff2);
	continue;
      }
      if (strstr (buff2, "attention") ||
          strstr (buff2, "error"))
         *needs_operator = TRUE;
      if (*needs_operator || strstr (buff2, "waiting") != NULL)
	strcpy (status, cp);
    }
    stat = PI_RES_OK;
    break;
  }
  pclose (p);
  return (stat);
}

/* 
 * pr_cancel
 *
 * cancel job id on printer pr on behalf of user
 */
pcrstat pr_cancel (char *pr, char *user, char *id)
{
  pcrstat stat = PC_RES_NO_SUCH_JOB;
  char cmdbuf[256];
  char resbuf[256];
  FILE *fd;
  int  i;

  pr = map_printer_name (pr);
  if (pr == NULL || suspicious(pr) || !valid_pr(pr))
     return (PC_RES_NO_SUCH_PRINTER);

  if (suspicious (id))
     return (PC_RES_NO_SUCH_JOB);

  sprintf (cmdbuf, "/usr/bin/lprm -P%s %s", pr, id);
  if ((fd = su_popen (user, cmdbuf, MAXTIME_FOR_CANCEL)) == NULL)
  {
    msg_out ("rpc.pcnfsd: su_popen failed");
    return (PC_RES_FAIL);
  }

  while (fgets (resbuf, 255, fd))
  {
    i = strlen (resbuf);
    if (i)
       resbuf[i-1] = '\0';   /* trim NL */
    if (strstr (resbuf, "dequeued") != NULL)
       stat = PC_RES_OK;
    if (strstr (resbuf, "unknown printer") != NULL)
       stat = PC_RES_NO_SUCH_PRINTER;
    if (strstr (resbuf, "Permission denied") != NULL)
       stat = PC_RES_NOT_OWNER;
  }
  if (su_pclose (fd) == 255)
     msg_out ("rpc.pcnfsd: su_pclose alert");
  return (stat);
}

/*
 * New subsystem here. We allow the administrator to define
 * up to NPRINTERDEFS aliases for printer names. This is done
 * using the "/etc/pcnfsd.conf" file, which is read at startup.
 * There are three entry points to this subsystem
 *
 * void add_printer_alias(char *printer, char *alias_for, char *command)
 *
 * This is invoked from "config_from_file()" for each
 * "printer" line. "printer" is the name of a printer; note that
 * it is possible to redefine an existing printer. "alias_for"
 * is the name of the underlying printer, used for queue listing
 * and other control functions. If it is "-", there is no
 * underlying printer, or the administrative functions are
 * not applicable to this printer. "command"
 * is the command which should be run (via "su_popen()") if a
 * job is printed on this printer. The following tokens may be
 * embedded in the command, and are substituted as follows:
 *
 * $FILE        -       path to the file containing the print data
 * $USER        -       login of user
 * $HOST        -       hostname from which job originated
 *
 * Tokens may occur multiple times. If The command includes no
 * $FILE token, the string " $FILE" is silently appended.
 *
 * pr_list list_virtual_printers()
 *
 * This is invoked from build_pr_list to generate a list of aliased
 * printers, so that the client that asks for a list of valid printers
 * will see these ones.
 *
 * char *map_printer_name (char *printer)
 *
 * If "printer" identifies an aliased printer, this function returns
 * the "alias_for" name, or NULL if the "alias_for" was given as "-".
 * Otherwise it returns its argument.
 *
 * char *expand_alias (char *printer, char *file, char *user, char *host)
 *
 * If "printer" is an aliased printer, this function returns a
 * pointer to a static string in which the corresponding command
 * has been expanded. Otherwise ot returns NULL.
 */

#define NPRINTERDEFS  128

int num_aliases = 0;

struct {
  char *a_printer;
  char *a_alias_for;
  char *a_command;
} alias [NPRINTERDEFS];


char default_cmd[] = "/usr/bin/lpr $FILE";

void add_printer_alias (char *printer, char *alias_for, char *command)
{
  if (alias_for == NULL || strlen(alias_for) == 0)
     alias_for = "lp";

  if (command == NULL || strlen (command) == 0)
     command = default_cmd;      /* see above */

  if (num_aliases < NPRINTERDEFS)
  {
    alias[num_aliases].a_printer   = my_strdup (printer);
    alias[num_aliases].a_alias_for = (strcmp(alias_for, "-") ?
                                      my_strdup(alias_for) : NULL);
    if (strstr (command, "$FILE"))
       alias[num_aliases].a_command = my_strdup (command);
    else
    {
      alias[num_aliases].a_command = (char*) grab (strlen(command)+8);
      strcpy (alias[num_aliases].a_command, command);
      strcat (alias[num_aliases].a_command, " $FILE");
    }
    num_aliases++;
  }
}


/* 
 * list_virtual_printers
 *
 * build a pr_list of all virtual printers (if any)
 */
pr_list list_virtual_printers (void)
{
  pr_list first = NULL;
  pr_list last = NULL;
  pr_list curr = NULL;
  int     i;

  if (num_aliases == 0)
     return (NULL);

  for (i = 0; i < num_aliases; i++)
  {
    curr = (struct pr_list_item*) grab (sizeof(struct pr_list_item));

    curr->pn = my_strdup (alias[i].a_printer);
    if (alias[i].a_alias_for == NULL)
         curr->device = my_strdup ("");
    else curr->device = my_strdup (alias[i].a_alias_for);
    curr->remhost = my_strdup ("");
    curr->cm      = my_strdup ("(alias)");
    curr->pr_next = NULL;
    if (last == NULL)
         first = curr;
    else last->pr_next = curr;
    last = curr;
  }
  return (first);
}


/* 
 * map_printer_name
 *
 * if a printer name is actually an alias, return the name of
 * the printer for which it is an alias; otherwise return the name
 */
char *map_printer_name (char *printer)
{
  int i;

  for (i = 0; i < num_aliases; i++)
  {
    if (!strcmp (printer, alias[i].a_printer))
       return (alias[i].a_alias_for);
  }
  return (printer);
}


/* 
 * substitute
 *
 * replace a token in a string as often as it occurs
 */
static void substitute (char *string, char *token, char *data)
{
  char temp[512];
  char *c;

  while ((c = strstr (string, token)) != NULL)
  {
    *c = '\0';
    strcpy (temp, string);
    strcat (temp, data);
    c += strlen (token);
    strcat (temp, c);
    strcpy (string, temp);
  }
}

/*
 * expand_alias
 *
 * expand an aliased printer command by substituting file/user/host
 */
char *expand_alias (char *printer, char *file, char *user, char *host)
{
  static char expansion[512];
  int i;

  for (i = 0; i < num_aliases; i++)
  {
    if (!strcmp (printer, alias[i].a_printer))
    {
      strcpy (expansion, alias[i].a_command);
      substitute (expansion, "$FILE", file);
      substitute (expansion, "$USER", user);
      substitute (expansion, "$HOST", host);
      return (expansion);
    }
  }
  return (NULL);
}
