/* 
 * RE_SID: @(%)/tmp_mnt/vol/dosnfs/shades_SCCS/unix/pcnfsd/v2/src/SCCS/
 * s.pcnfsd_test.c 1.4 93/01/29 17:53:37 SMI
 */

#include <stdio.h>
#include <stdlib.h>
#include <rpc/rpc.h>
#include <tcp.h>

#include "pcnfsd.h"
#include "extern.h"

CLIENT *cl;
CLIENT *cl2;
char   *server;
char    spooldirbuff[256];
char    filenamebuff[256];
char    last_id[32] = "";

int  test_v2_info  (void);
int  test_v2_list  (void);
int  test_v2_auth  (char *host_name, char *user_name, char *pwrd);
int  test_v2_init  (char *host_name, char *printer);
int  test_v2_start (char *host_name, char *printer, char *user_name, char *tag1, char *tag2);
int  test_v2_queue (char *printer, char *user_name, int private);
int  test_v2_cancel(char *host_name, char *printer, char *user_name, char *id);
int  test_v2_stat  (char *printer);
int  test_v2_map   (void);
void good          (void);
void bad           (char *reason);

int main (int argc, char **argv)
{
  char *host_name;
  char *printer;
  char *user_name;
  char *passwd;
  char *transport = "udp";

  if (argc < 6 || argc > 7)
  {
    fprintf (stderr, "usage: %s server host printer user password [transport]\n",
	     argv[0]);
    return (1);
  }

  server    = argv[1];
  host_name = argv[2];
  printer   = argv[3];
  user_name = argv[4];
  passwd    = argv[5];
  if (argc == 7)
    transport = argv[6];

  dbug_init();
//rpc_init();
  sock_init();

  cl = clnt_create (server, PCNFSDPROG, PCNFSDVERS, transport);
  if (cl == NULL)
  {
    clnt_pcreateerror (server);
    return (1);
  }
  cl2 = clnt_create (server, PCNFSDPROG, PCNFSDV2, transport);
  if (cl2 == NULL)
  {
    clnt_pcreateerror (server);
    return (1);
  }

  good();
  test_v2_info();
  good();
  test_v2_auth (host_name, user_name, passwd);
  bad ("Invalid password");
  test_v2_auth (host_name, user_name, "bogus");
  good();
  test_v2_list();
  good();
  test_v2_init (host_name, printer);
  good();
  test_v2_start (host_name, printer, user_name, "foo", "foo");
  good();
  test_v2_start (host_name, printer, user_name, "bar", "bar");
  bad ("No such file to print");
  test_v2_start (host_name, printer, user_name, "bletch", "gack");
  good();
  test_v2_queue (printer, user_name, FALSE);

  if (strlen(last_id))
  {
    bad ("Cancelling job with bad username");
    test_v2_cancel (host_name, printer, "nosuchuser", last_id);
    good();
    test_v2_cancel (host_name, printer, user_name, last_id);
  }
  bad ("Cancelling unknown job");
  test_v2_cancel (host_name, printer, user_name, "99999");
  bad ("Cancelling job on invalid printer");
  test_v2_cancel (host_name, "nosuchprinter", user_name, last_id);
  good();
  test_v2_queue (printer, user_name, TRUE);
  bad ("Checking queue on invalid printer");
  test_v2_queue ("nosuchprinter", user_name, TRUE);
  good();
  test_v2_stat (printer);
  bad ("Checking status of invalid printer");
  test_v2_stat ("nosuchprinter");
  good();
  test_v2_map();
  return (0);
}

int test_v2_info (void)
{
  v2_info_args     a;
  v2_info_results *rp;
  long *gp;
  int   i;

  a.vers = "Sun Microsystems PCNFSD test subsystem V1";
  a.cm   = "-";
  printf ("\ninvoking pr_info_2\n");

  rp = pcnfsd2_info_2 (&a, cl2);
  if (!rp)
  {
    clnt_perror (cl2, server);
    return (1);
  }

  printf ("results: vers = '%s', cm = '%s'\n",
	  rp->vers, rp->cm);
  printf ("facilities_len = %ld\n", rp->facilities.facilities_len);

  if (rp->facilities.facilities_len)
  {
    gp = rp->facilities.facilities_val;
    for (i = 0; i < rp->facilities.facilities_len; i++)
      printf (" procedure %2d: %6ld\n", i, *gp++);
    printf ("\n");
  }

  /* free up allocated strings
   */
  if (rp->cm)
     free (rp->cm);
  if (rp->facilities.facilities_val)
     free (rp->facilities.facilities_val);
  if (rp->vers)
     free (rp->vers);
  return (0);
}

int test_v2_auth (char *host_name, char *user_name, char *pwrd)
{
  v2_auth_args a;
  v2_auth_results *rp;
  char   uname[32];
  char   pw[64];

  scramble (user_name, uname);
  scramble (pwrd, pw);
  a.system = host_name;
  a.id     = uname;
  a.pw     = pw;
  a.cm     = "-";
  printf ("\ninvoking pr_auth_2\n");

  rp = pcnfsd2_auth_2 (&a, cl2);
  if (!rp)
  {
    clnt_perror (cl2, server);
    return (1);
  }

  if (rp->stat == AUTH_RES_FAIL)
    printf ("results: stat = AUTH_RES_FAIL\n");
  else
  {
    printf ("results: stat = %d, uid = %lu, gid = %lu,\n "
            "homedir= '%s', cm = '%s'\n",
	    rp->stat, rp->uid, rp->gid, rp->home, rp->cm);
    printf ("gids_len = %ld", rp->gids.gids_len);
    if (rp->gids.gids_len)
    {
      u_long *gp = rp->gids.gids_val;
      int     i;

      for (i = 0; i < rp->gids.gids_len; i++)
        printf (" %lu", *gp++);
      printf ("\n");
    }
  }
 /* free up allocated strings
  */
  if (rp->cm)
     free (rp->cm);
  if (rp->gids.gids_val)
     free (rp->gids.gids_val);
  if (rp->home)
     free (rp->home);
  return (0);
}

int test_v2_init (char *host_name, char *printer)
{
  v2_pr_init_args     a;
  v2_pr_init_results *rp;

  a.system = host_name;
  a.pn     = printer;
  a.cm     = "-";
  printf ("\ninvoking pr_init_2\n");

  rp = pcnfsd2_pr_init_2 (&a, cl2);
  if (!rp)
  {
    clnt_perror (cl2, server);
    return (1);
  }
  printf ("results: stat = %d, dir = '%s', cm = '%s'\n",
	  rp->stat, rp->dir, rp->cm);
  strcpy (spooldirbuff, rp->dir);

  if (rp->cm)
     free (rp->cm);
  if (rp->dir)
     free (rp->dir);
  return (0);
}


int test_v2_start (char *host_name, char *printer, char *user_name,
		   char *tag1, char *tag2)
{
  v2_pr_start_args     a;
  v2_pr_start_results *rp;
  FILE  *fp;

  printf ("\ntesting start print v2\n");

  if (strcmp(server, "localhost"))
  {
    printf ("sorry - can only test start print on 'localhost'\n");
    return (1);
  }

  sprintf (filenamebuff, "%s/%s", spooldirbuff, tag1);

  fp = fopen (filenamebuff, "w");
  if (!fp)
  {
    perror ("creating test file");
    return (1);
  }
  fputs ("foo bar bletch\n", fp);
  fclose (fp);

  a.system = host_name;
  a.pn     = printer;
  a.user   = user_name;
  a.file   = tag2;
  a.opts   = "xxxx";
  a.copies = 1;
  a.cm     = "-";

  printf ("\ninvoking pr_start_2\n");

  rp = pcnfsd2_pr_start_2 (&a, cl2);
  if (!rp)
  {
    clnt_perror (cl2, server);
    return (1);
  }
  printf ("results: stat = %d, jobid = '%s', cm = '%s'\n",
	  rp->stat, rp->id, rp->cm);

  if (rp->stat == PS_RES_OK)
     strcpy (last_id, rp->id);

  if (rp->cm)
     free (rp->cm);
  if (rp->id)
     free (rp->id);
  return (0);
}


int test_v2_cancel (char *host_name, char *printer, char *user_name, char *id)
{
  v2_pr_cancel_args     a;
  v2_pr_cancel_results *rp;

  printf ("\ntesting cancel print v2\n");

  a.system = host_name;
  a.pn     = printer;
  a.user   = user_name;
  a.id     = id;
  a.cm     = "-";

  printf ("\ninvoking pr_cancel_2 for job %s on printer %s\n",
	  id, printer);

  rp = pcnfsd2_pr_cancel_2 (&a, cl2);
  if (!rp)
  {
    clnt_perror (cl2, server);
    return (1);
  }
  printf ("results: stat = %d, cm = '%s'\n",
	  rp->stat, rp->cm);

  if (rp->cm)
     free (rp->cm);
  return (0);
}

int test_v2_list (void)
{
  v2_pr_list_results *rp;
  pr_list             curr;
  char                a;

  printf ("\ninvoking pr_list_2\n");

  rp = pcnfsd2_pr_list_2 (&a, cl2);
  if (!rp)
  {
    clnt_perror (cl2, server);
    return (1);
  }
  printf ("results: cm = '%s', printerlist:\n", rp->cm);

  for (curr = rp->printers; curr; curr = curr->pr_next)
  {
    printf ("  name '%s' ", curr->pn);
    if (strlen(curr->remhost) > 0)
         printf ("remote: srvr '%s', name '%s'",
                  curr->remhost, curr->device);
    else printf ("local device = '%s'", curr->device);
    printf (", cm = '%s'\n", curr->cm);
  }
  printf ("end of list\n");

  if (rp->cm)
     free (rp->cm);
  if (rp->printers)
  {
    printf ("freeing results\n");
    free_pr_list_item (rp->printers);
  }
  return (0);
}


int test_v2_queue (char *printer, char *user_name, int private)
{
  struct v2_pr_queue_args a;
  v2_pr_queue_results    *rp;
  pr_queue                curr;

  a.pn        = printer;
  a.system    = "foo";
  a.user      = user_name;
  a.just_mine = private;
  a.cm        = "no";

  printf ("\ninvoking pr_queue_2 (just_mine = %d)\n", private);

  rp = pcnfsd2_pr_queue_2 (&a, cl2);
  if (!rp)
  {
    clnt_perror (cl2, server);
    return (1);
  }
  printf ("results: stat = %d, qlen = %ld, qshown = %ld cm = '%s', queue:\n",
	  rp->stat, rp->qlen, rp->qshown, rp->cm);

  for (curr = rp->jobs; curr; curr = curr->pr_next)
  {
    printf ("rank = %2ld, id = '%s', size = '%s', status = '%s'\n",
            curr->position, curr->id, curr->size, curr->status);
    printf ("            user = '%s', file = '%s', cm = '%s'\n",
            curr->user, curr->file, curr->cm);
  }
  printf ("end of list\n");

  if (rp->cm)
     free (rp->cm);
  if (rp->jobs)
  {
    printf ("freeing results\n");
    free_pr_queue_item (rp->jobs);
  }
  return (0);
}
 

int test_v2_stat (char *printer)
{
  v2_pr_status_args     a;
  v2_pr_status_results *rp;

  printf ("\ntesting status print v2\n");

  a.pn = printer;
  a.cm = "-";

  printf ("\ninvoking pr_status_2\n");

  rp = pcnfsd2_pr_status_2 (&a, cl2);
  if (!rp)
  {
    clnt_perror (cl2, server);
    return (1);
  }
  printf ("results: stat = %d, cm = '%s'\n", rp->stat, rp->cm);
  if (rp->stat == PI_RES_OK)
  {
    printf ("avail = %s, ", (rp->avail ? "YES" : "NO"));
    printf ("printing = %s, ", (rp->printing ? "YES" : "NO"));
    printf ("needs_operator = %s, ", (rp->needs_operator ? "YES" : "NO"));
    printf ("qlen = %ld, status = '%s'\n", rp->qlen, rp->status);
  }

  if (rp->cm)
     free (rp->cm);
  if (rp->status)
     free (rp->status);
  return (0);
}

struct mapreq_arg_item *make_mapreq_entry (mapreq t, int i, char *n,
                                           struct mapreq_arg_item *next)
{
  struct mapreq_arg_item *x = malloc (sizeof(*x));

  if (!x)
  {
    fprintf (stderr, "out of memory\n");
    exit (-1);
  }
  x->req  = t;
  x->id   = i;
  x->name = (n ? n : "");
  x->mapreq_next = next;
  return (x);
}

int test_v2_map (void)
{
  v2_mapid_args          a;
  v2_mapid_results       *rp;
  struct mapreq_res_item *rip;

  a.cm = "-";
  a.req_list = make_mapreq_entry (MAP_REQ_UID, 906, NULL,
               make_mapreq_entry (MAP_REQ_GID,   1, NULL,
               make_mapreq_entry (MAP_REQ_UNAME, 0, "root",
               make_mapreq_entry (MAP_REQ_GNAME, 0, "wheel",
               make_mapreq_entry (MAP_REQ_UNAME, 0, "bogus", NULL)))));

  printf ("\ninvoking pr_mapid_2\n");

  rp = pcnfsd2_mapid_2 (&a, cl2);
  if (!rp)
  {
    clnt_perror (cl2, server);
    return (1);
  }
  printf ("results: cm = '%s', result list %s\n",
	  rp->cm, rp->res_list ? "follows" : "omitted");

  for (rip = rp->res_list; rip; rip = rip->mapreq_next)
      printf ("request type = %d, status = %d, id = %ld, name = '%s'\n",
              rip->req, rip->stat, rip->id, rip->name);

  /* XXX should free up results */
  return (0);
}


void good (void)
{
  printf ("\n");
  printf ("********************************************************\n");
  printf ("********************************************************\n");
  printf ("**      The following test is expected to SUCCEED     **\n");
  printf ("********************************************************\n");
  printf ("********************************************************\n");
}

void bad (char *reason)
{
  printf ("\n");
  printf ("********************************************************\n");
  printf ("********************************************************\n");
  printf ("**      The following test is expected to FAIL        **\n");
  printf ("**                    Reason:                         **\n");
  printf ("** %50s **\n", reason);
  printf ("********************************************************\n");
  printf ("********************************************************\n");
}


