/* @(#)rpc_clntout.c  2.1 88/08/01 4.0 RPCSRC */
/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

/*
 * rpc_clntout.c, Client-stub outputter for the RPC protocol compiler
 * Copyright (C) 1987, Sun Microsytsems, Inc.
 */

#include <stdio.h>
#include <string.h>

#include "rpc_scan.h"
#include "rpc_pars.h"
#include "rpc_util.h"

#define DEFAULT_TIMEOUT 25  /* in seconds */

static char *ampr (char *type)
{
  if (isvectordef(type, REL_ALIAS))
     return ("");
  return ("&");
}

static void printbody (proc_list *proc)
{
  fprintf (fout, "  static ");
  if (streq(proc->res_type, "void"))
       fprintf (fout, "char ");
  else ptype (proc->res_prefix, proc->res_type, 0);

  fprintf (fout, "res;\n\n");
  fprintf (fout, "  memset ((void*)%sres,0,sizeof(res));\n",
           ampr(proc->res_type));

  fprintf (fout,
           "  if (clnt_call(clnt,%s,\n"
           "       (xdrproc_t) xdr_%s, (caddr_t) argp,\n"
           "       (xdrproc_t) xdr_%s, (caddr_t) %sres, TIMEOUT) != RPC_SUCCESS)\n"
           "     return (NULL);\n",
           proc->proc_name, stringfix(proc->arg_type),
           stringfix(proc->res_type), ampr(proc->res_type));

  if (streq(proc->res_type, "void"))
       fprintf (fout, "  return ((void *)%sres);\n", ampr(proc->res_type));
  else fprintf (fout, "  return (%sres);\n",         ampr(proc->res_type));

}

static void write_program (definition *def)
{
  version_list *vp;
  proc_list    *proc;

  for (vp = def->def.pr.versions; vp; vp = vp->next)
      for (proc = vp->procs; proc; proc = proc->next)
      {
        fprintf (fout, "\n");
        ptype (proc->res_prefix, proc->res_type, 1);
        fprintf (fout, "*");
        pvname (proc->proc_name, vp->vers_num);
        fprintf (fout, " (");
        ptype (proc->arg_prefix, proc->arg_type, 1);
        fprintf (fout,"*argp, CLIENT *clnt)\n{\n");
        printbody (proc);
        fprintf (fout, "}\n\n");
      }
}

void write_stubs (void)
{
  list       *l;
  definition *def;

  fprintf (fout, "\n/* Default timeout can be changed using clnt_control() */\n");
  fprintf (fout, "static struct timeval TIMEOUT = { %d, 0 };\n",
           DEFAULT_TIMEOUT);
  for (l = defined; l; l = l->next)
  {
    def = (definition *) l->val;
    if (def->def_kind == DEF_PROGRAM)
       write_program (def);
  }
}

