/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */
#if !defined(lint) && defined(SCCSIDS)
static char sccsid[] = "@(#)pmap_rmt.c 1.21 87/08/27 Copyr 1984 Sun Micro";
#endif

/*
 * pmap_rmt.c
 * Client interface to pmap rpc service.
 * remote call and broadcast service
 *
 * Copyright (C) 1984, Sun Microsystems, Inc.
 */

#include <stdio.h>
#include <errno.h>
#include <process.h>
#include <rpc/rpc.h>
#include <rpc/pmap_pro.h>
#include <rpc/pmap_cln.h>
#include <rpc/pmap_rmt.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#ifdef __DJGPP__
#include <unistd.h>
#endif

#include "rpc_mac.h"

typedef bool_t (*resultproc_t)();

static struct timeval timeout = { 3, 0 };


/*
 * pmapper remote-call-service interface.
 * This routine is used to call the pmapper remote call service
 * which will look up a service program in the port maps, and then
 * remotely call that routine with the given parameters.  This allows
 * programs to do a lookup and call in one step.
*/
enum clnt_stat pmap_rmtcall (addr, prog, vers, proc, xdrargs,
                             argsp, xdrres, resp, tout, port_ptr)
  struct    sockaddr_in *addr;
  u_long    prog, vers, proc;
  xdrproc_t xdrargs, xdrres;
  caddr_t   argsp, resp;
  struct    timeval tout;
  u_long   *port_ptr;
{
  CLIENT *client;
  struct rmtcallargs a;
  struct rmtcallres  r;
  enum   clnt_stat stat;
#ifdef __MSDOS__
  udp_Socket *socket = NULL;
#else
  int socket = -1;
#endif

  addr->sin_port = htons (PMAPPORT);
  client = clntudp_create (addr, PMAPPROG, PMAPVERS, timeout, (int*)&socket);
  if (client)
  {
    a.prog        = prog;
    a.vers        = vers;
    a.proc        = proc;
    a.args_ptr    = argsp;
    a.xdr_args    = xdrargs;
    r.port_ptr    = port_ptr;
    r.results_ptr = resp;
    r.xdr_results = xdrres;
    stat = CLNT_CALL (client, PMAPPROC_CALLIT, xdr_rmtcall_args,
                      &a, xdr_rmtcallres, &r, tout);
    CLNT_DESTROY (client);
  }
  else
    stat = RPC_FAILED;

#ifdef __MSDOS__
  sock_close (socket);
#else
  close (socket);
#endif

  addr->sin_port = 0;
  return (stat);
}


/*
 * XDR remote call arguments
 * written for XDR_ENCODE direction only
 */
bool_t xdr_rmtcall_args (XDR *xdrs, struct rmtcallargs *cap)
{
  u_int lenposition, argposition, position;

  if (xdr_u_long(xdrs, &cap->prog) &&
      xdr_u_long(xdrs, &cap->vers) &&
      xdr_u_long(xdrs, &cap->proc))
  {
    lenposition = XDR_GETPOS (xdrs);
    if (!xdr_u_long(xdrs, &cap->arglen))
       return (FALSE);

    argposition = XDR_GETPOS (xdrs);
    if (!(*(cap->xdr_args))(xdrs,cap->args_ptr))
       return (FALSE);

    position = XDR_GETPOS (xdrs);
    cap->arglen = (u_long)position - (u_long)argposition;
    XDR_SETPOS (xdrs, lenposition);
    if (!xdr_u_long(xdrs, &cap->arglen))
       return (FALSE);

    XDR_SETPOS (xdrs, position);
    return (TRUE);
  }
  return (FALSE);
}

/*
 * XDR remote call results
 * written for XDR_DECODE direction only
 */
bool_t xdr_rmtcallres (XDR *xdrs, struct rmtcallres *crp)
{
  caddr_t port_ptr = (caddr_t)crp->port_ptr;

  if (xdr_reference(xdrs,&port_ptr,sizeof(u_long),xdr_u_long) &&
      xdr_u_long(xdrs, &crp->resultslen))
  {
    crp->port_ptr = (u_long*)port_ptr;
    return (*(crp->xdr_results))(xdrs, crp->results_ptr);
  }
  return (FALSE);
}


/*
 * return both the limited and direct broadcast addresses
 */
static int getbroadcastnets (struct in_addr *addrs)
{
#if 0
  struct in_addr my_addr;

  my_addr.s_addr  = gethostid();
  addrs[0]        = inet_makeaddr (inet_netof(my_addr),INADDR_ANY);
  addrs[1].s_addr = INADDR_BROADCAST;

  if (addrs[0].s_addr == addrs[1].s_addr)
     return (1);
  return (2);
#else
  addrs[0].s_addr = INADDR_BROADCAST;
  return (1);
#endif
}

/*
 *  NB !!: this function uses ~13000 bytes of stack
 */
enum clnt_stat
clnt_broadcast (prog, vers, proc, xargs, argsp, xresults, resultsp, eachresult)
  u_long       prog;        /* program number */
  u_long       vers;        /* version number */
  u_long       proc;        /* procedure number */
  xdrproc_t    xargs;       /* xdr routine for args */
  char        *argsp;       /* pointer to args */
  xdrproc_t    xresults;    /* xdr routine for results */
  char        *resultsp;    /* pointer to results */
  resultproc_t eachresult;  /* call with each result obtained */
{
  enum   clnt_stat stat;
  AUTH  *unix_auth         = authunix_create_default();
  XDR    xdr_stream, *xdrs = &xdr_stream;
  int    i, outlen, inlen  = 0;
  u_long xid, port;
  struct in_addr     addrs[20];
  struct sockaddr_in raddr;     /* response address */
  struct rmtcallargs a;
  struct rmtcallres  r;
  struct rpc_msg     msg;
  struct timeval     t;
  char   outbuf [MAX_BROADCAST_SIZE];
  char   inbuf  [UDPMSGSIZE];
  bool_t done = FALSE;

#ifdef __MSDOS__
  udp_Socket sock;
  int        status;
  int        nets = getbroadcastnets (addrs);

#else
  struct sockaddr_in baddr;    /* broadcast address */
  fd_set mask, readfds;
  int    nets, fromlen;
  int    sock = -1;

  /*
   * initialization: create a socket, a broadcast address, and
   * preserialize the arguments into a send buffer.
   */
  if ((sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
  {
    perror ("Cannot create socket for broadcast rpc");
    stat = RPC_CANTSEND;
    goto done_broad;
  }

  {
    int on = 1;
    if (setsockopt(sock,SOL_SOCKET,SO_BROADCAST,&on,sizeof(on)) < 0)
    {
      perror ("Cannot set socket option SO_BROADCAST");
      stat = RPC_CANTSEND;
      goto done_broad;
    }
  }

  FD_ZERO (&mask);
  FD_SET (sock, &mask);
  nets = getbroadcastnets (addrs);
  memset (&baddr,0,sizeof(baddr));
  baddr.sin_family      = AF_INET;
  baddr.sin_port        = htons (PMAPPORT);
  baddr.sin_addr.s_addr = htonl (INADDR_ANY);
#endif

  gettimeofday (&t,NULL);
  msg.rm_xid = xid = getpid() ^ t.tv_sec ^ t.tv_usec;
  t.tv_usec = 0;
  msg.rm_direction       = CALL;
  msg.rm_call.cb_rpcvers = RPC_MSG_VERSION;
  msg.rm_call.cb_prog    = PMAPPROG;
  msg.rm_call.cb_vers    = PMAPVERS;
  msg.rm_call.cb_proc    = PMAPPROC_CALLIT;
  msg.rm_call.cb_cred    = unix_auth->ah_cred;
  msg.rm_call.cb_verf    = unix_auth->ah_verf;
  a.prog        = prog;
  a.vers        = vers;
  a.proc        = proc;
  a.xdr_args    = xargs;
  a.args_ptr    = (caddr_t) argsp;
  r.port_ptr    = &port;
  r.xdr_results = xresults;
  r.results_ptr = (caddr_t) resultsp;
  xdrmem_create (xdrs, outbuf, sizeof(outbuf), XDR_ENCODE);

  if (!xdr_callmsg(xdrs,&msg) || !xdr_rmtcall_args(xdrs,&a))
  {
    stat = RPC_CANTENCODEARGS;
    goto done_broad;
  }
  outlen = xdr_getpos (xdrs);
  xdr_destroy (xdrs);

  MSG ("listening for portmappers.");

  /*
   * Basic loop: broadcast a packet and wait a while for response(s).
   * The response timeout grows larger per iteration.
   */
  for (t.tv_sec = 4; t.tv_sec <= 14; t.tv_sec += 2)
  {
    for (i = 0; i < nets; i++)
    {
#ifdef __MSDOS__
      udp_open (&sock, 0, ntohl(addrs[i].s_addr), PMAPPORT, NULL);
      sock_write (&sock, outbuf, outlen);
#else
      baddr.sin_addr = addrs[i];
      if (sendto(sock, outbuf, outlen, 0, (struct sockaddr *)&baddr,
                sizeof (struct sockaddr)) != outlen)
      {
        perror ("Cannot send broadcast packet");
        stat = RPC_CANTSEND;
        goto done_broad;
      }
#endif
    }
    if (eachresult == NULL)
    {
      stat = RPC_SUCCESS;
      goto done_broad;
    }

  recv_again:

    msg.acpted_rply.ar_verf          = _null_auth;
    msg.acpted_rply.ar_results.where = (caddr_t)&r;
    msg.acpted_rply.ar_results.proc  = xdr_rmtcallres;

#ifdef __MSDOS__
    status = 0;
    sock_wait_input (&sock, t.tv_sec, NULL, &status);
    inlen = sock_fastread (&sock, (BYTE*)&inbuf, UDPMSGSIZE);
    if (inlen < sizeof(u_long))
       goto recv_again;

  sock_err:
    switch (status)
    {
      case 0:
           _getpeername ((const tcp_Socket*)&sock, (void*)&raddr, NULL);
           raddr.sin_addr.s_addr = htonl (raddr.sin_addr.s_addr);
           break;  /* okay */
      case 1:
           MSG1 ("PMAP Broadcast: ", sockerr((tcp_Socket*)&sock));
           stat = RPC_CANTRECV;
           goto done_broad;
      case -1:
           MSG1 ("PMAP Broadcast (timeout): ", sockerr((tcp_Socket*)&sock));
           stat = RPC_TIMEDOUT;
           continue;
    }

#else
    readfds = mask;
    switch (select(_rpc_dtablesize(),&readfds,NULL,NULL,&t))
    {
      case 0:  /* timed out */
           stat = RPC_TIMEDOUT;
           continue;

      case -1:  /* some kind of error */
           if (errno == EINTR)
              goto recv_again;
           perror ("Broadcast select problem");
           stat = RPC_CANTRECV;
           goto done_broad;
    }

  try_again:
    fromlen = sizeof (struct sockaddr);
    inlen   = recvfrom (sock, inbuf, UDPMSGSIZE, 0,
                        (struct sockaddr*)&raddr, &fromlen);
    if (inlen < 0)
    {
      if (errno == EINTR)
         goto try_again;
      perror ("Cannot receive reply to broadcast");
      stat = RPC_CANTRECV;
      goto done_broad;
    }
    if (inlen < sizeof(u_long))
       goto recv_again;
#endif


    /*
     * see if reply transaction id matches sent id.
     * If so, decode the results.
     */
    xdrmem_create (xdrs, inbuf, inlen, XDR_DECODE);
    if (xdr_replymsg(xdrs, &msg))
    {
      if ((msg.rm_xid              == xid)          &&
          (msg.rm_reply.rp_stat    == MSG_ACCEPTED) &&
          (msg.acpted_rply.ar_stat == SUCCESS))
      {
        raddr.sin_port = htons (port);
        done = (*eachresult) (resultsp, &raddr);
      }
    }
    else
    {
#if 1
      /* some kind of deserialization problem ... */
      if (msg.rm_xid == xid)
         MSG ("Broadcast deserialization problem");
      /* otherwise, just random garbage */
#endif
    }

    xdrs->x_op = XDR_FREE;
    msg.acpted_rply.ar_results.proc = xdr_void;
    xdr_replymsg (xdrs, &msg);
    (*xresults) (xdrs, resultsp);
    xdr_destroy (xdrs);
    if (done)
    {
      stat = RPC_SUCCESS;
      goto done_broad;
    }
    else
      goto recv_again;
  }

done_broad:

#ifdef __MSDOS__
  sock_close (&sock);
#else
  if (sock != -1)
     close (sock);
#endif

  AUTH_DESTROY (unix_auth);
  return (stat);
}

