/*
 * Copyright (c) 1984 by Sun Microsystems, Inc.
 *
 * pmap_loc.c (originally /etc/portmap.c),
 * Implements the program,version to port number mapping for rpc.
 *
 * Adapted for Waterloo TCP/IP by G.Vanem <giva@bgnett.no>
 * Only installs UDP service
 */

/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

#include <stdio.h>
#include <malloc.h>
#include <rpc/rpc.h>
#include <rpc/pmap_pro.h>
#include <rpc/pmap_cln.h>
#include <rpc/svc.h>
#include <sys/socket.h>

#include "rpc_mac.h"

static struct pmaplist *_pmaplist = NULL;

static int      pmap_started = 0;
static int      pmap_stopped = 0;
static int      pmap_sock    = -1;
static SVCXPRT *pmap_xprt;

static void pmap_daemon (void);
static void reg_service (struct svc_req *rqstp, SVCXPRT *xprt);

/*
 *  Create a Portmapper service on UDP port 111
 */
int pmap_start_server (void)
{
  struct pmaplist   *pml;
  struct sockaddr_in addr;
  int    len = sizeof(struct sockaddr_in);

  if (pmap_started)
     return (1);

  if ((pmap_sock = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP)) < 0)
     return (0);

#if 1
  {
    int debug = 1;
    if (setsockopt(pmap_sock,SOL_SOCKET,SO_DEBUG,&debug,sizeof(debug)) < 0)
       return (0);
  }
#endif

  addr.sin_addr.s_addr = INADDR_ANY;
  addr.sin_family      = AF_INET;
  addr.sin_port        = htons (PMAPPORT);
  if (bind(pmap_sock, (struct sockaddr *)&addr,len) != 0)
     return (0);

  if ((pmap_xprt = svcudp_create(pmap_sock)) == NULL)
     return (0);

  /* make an entry for ourself
   */
  pml = malloc (sizeof(*pml));
  pml->pml_next        = NULL;
  pml->pml_map.pm_prog = PMAPPROG;
  pml->pml_map.pm_vers = PMAPVERS;
  pml->pml_map.pm_prot = IPPROTO_UDP;
  pml->pml_map.pm_port = PMAPPORT;
  _pmaplist = pml;

  svc_register (pmap_xprt,PMAPPROG,PMAPVERS,reg_service,FALSE);
  addwattcpd (pmap_daemon);
  pmap_started = 1;
  atexit ((void(*))pmap_stop_server);

#ifdef DEBUG
  fprintf (stderr, "pmap_start_server(), sock = %d\n",pmap_sock); // !!
#endif
  return (1);
}

int pmap_stop_server (void)
{
  if (!pmap_stopped)
  {
    SVC_DESTROY (pmap_xprt);
    delwattcpd (pmap_daemon);
    pmap_stopped = 1;
  }
  return (1);
}

static void pmap_daemon (void)
{
  struct timeval timeout;
  int    rc;

  timeout.tv_sec  = 0;
  timeout.tv_usec = 0;  /* immediate return */

  if ((rc = select_s(_rpc_dtablesize(),&svc_fdset,NULL,NULL,&timeout)) > 0)
  {
    fprintf (stderr, "pmap_daemon(), rc = %d\n", rc); // !!
    svc_getreqset (&svc_fdset);
  }
}


static struct pmaplist *find_service (u_long prog, u_long vers, int prot)
{
  struct pmaplist *hit = NULL;
  struct pmaplist *pml;

  for (pml = _pmaplist; pml; pml = pml->pml_next)
  {
    if (pml->pml_map.pm_prog != prog ||
        pml->pml_map.pm_prot != prot)
       continue;
    hit = pml;
    if (pml->pml_map.pm_vers == vers)
       break;
  }
  return (hit);
}

/* 
 * 1 OK, 0 not
 */
static void reg_service (struct svc_req *rqstp, SVCXPRT *xprt)
{
  struct pmap      reg;
  struct pmaplist *pml, *prevpml, *fnd;
  int    ans, port;
  void  *t;
  
#ifdef DEBUG
  fprintf (stderr, "server: about do a switch\n");
#endif

  switch (rqstp->rq_proc)
  {
    case PMAPPROC_NULL:
         /*
          * Null proc call
          */
         svc_sendreply (xprt,xdr_void,NULL);
         break;

    case PMAPPROC_SET:
         /*
          * Set a program,version to port mapping
          */
         if (!svc_getargs(xprt,xdr_pmap,&reg))
            svcerr_decode (xprt);
         else
         {
           /*
            * check to see if already used.
            * find_service() returns a hit even if
            * the versions don't match, so check for it
            */
           fnd = find_service (reg.pm_prog, reg.pm_vers, reg.pm_prot);
           if (fnd && fnd->pml_map.pm_vers == reg.pm_vers)
           {
             if (fnd->pml_map.pm_port == reg.pm_port)
             {
               ans = 1;
               goto done;
             }
             else
             {
               ans = 0;
               goto done;
             }
           }
           else
           {
             /*
              * add to END of list
              */
             pml = malloc (sizeof(*pml));
             pml->pml_map  = reg;
             pml->pml_next = NULL;
             if (_pmaplist == NULL)
                 _pmaplist = pml;
             else
             {
               for (fnd = _pmaplist; fnd->pml_next; fnd = fnd->pml_next)
                    ;
               fnd->pml_next = pml;
             }
             ans = 1;
           }
         done:
           svc_sendreply (xprt,xdr_long,(char*)&ans);
         }
         break;

    case PMAPPROC_UNSET:
         /*
          * Remove a program,version to port mapping.
          */
         if (!svc_getargs(xprt, xdr_pmap, &reg))
            svcerr_decode (xprt);
         else
         {
           ans = 0;
           for (prevpml = NULL, pml = _pmaplist; pml; )
           {
             if (pml->pml_map.pm_prog != reg.pm_prog ||
                 pml->pml_map.pm_vers != reg.pm_vers)
             {
               /* both pml & prevpml move forwards
                */
               prevpml = pml;
               pml = pml->pml_next;
               continue;
             }
             /* found it; pml moves forward, prevpml stays
              */
             ans = 1;
             t   = (void*) pml;
             pml = pml->pml_next;
             if (!prevpml)
                  _pmaplist = pml;
             else prevpml->pml_next = pml;
             free (t);
           }
           svc_sendreply (xprt,xdr_long,(char*)&ans);
         }
         break;

    case PMAPPROC_GETPORT:
         /*
          * Lookup the mapping for a program,version and return its port
          */
         if (!svc_getargs(xprt,xdr_pmap,&reg))
            svcerr_decode (xprt);
         else
         {
           fnd = find_service (reg.pm_prog, reg.pm_vers, reg.pm_prot);
           if (fnd)
                port = fnd->pml_map.pm_port;
           else port = 0;

           svc_sendreply (xprt,xdr_long,(char*)&port);
         }
         break;

    case PMAPPROC_DUMP:
         /*
          * Return the current set of mapped program,version
          */
         if (!svc_getargs(xprt,xdr_void,NULL))
              svcerr_decode (xprt);
         else svc_sendreply (xprt,xdr_pmaplist,(char*)&_pmaplist);
         break;

    case PMAPPROC_CALLIT:
#if 0
         /*
          * Calls a procedure on the local machine.  If the requested
          * procedure is not registered this procedure does not return
          * error information!!
          * This procedure is only supported on rpc/udp and calls via
          * rpc/udp.  It passes null authentication parameters.
          */
         callit (rqstp,xprt);
#endif
         break;

    default:
         svcerr_noproc (xprt);
         break;
  }
}


