/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

/*
 * Copyright (c) 1988 by Sun Microsystems, Inc.
 */

/*
 * rtime - get time from remote machine
 *
 * gets time, obtaining value from host
 * on the udp/time socket.  Since timeserver returns
 * with time of day in seconds since Jan 1, 1900,  must
 * subtract seconds before Jan 1, 1970 to get
 * what unix uses.
 */

#include <stdio.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "rpc_mac.h"


#define NYEARS   (1970 - 1900)
#define TOFFSET  (time_t)(60*60*24*(365*NYEARS + (NYEARS/4)))

#ifndef __MSDOS__
static void do_close (int s)
{
  int save = errno;
  close (s);
  errno = save;
}
#endif

int rtime (struct sockaddr_in *addrp, struct timeval *timep, struct timeval *timeout)
{
#ifdef __MSDOS__
  u_long     thetime;
  udp_Socket sock;
  int        len;
  u_short    port = _rpc_cfg.rtime_port;
  u_long     host = ntohl (addrp->sin_addr.s_addr);

  udp_open (&sock, 0, host, port, NULL);
  sock_write (&sock, "\n",1);
  sock_wait_input (&sock, timeout->tv_sec, NULL, NULL);

  len = sock_read (&sock, (char*)&thetime, sizeof(thetime));
  if (len < sizeof(thetime))
  {
    errno = EIO;
    return (-1);
  }
  thetime = ntohl (thetime);
  timep->tv_sec  = (time_t)thetime - TOFFSET;
  timep->tv_usec = 0;
  sock_close (&sock);
  return (0);

sock_err:
  sock_close (&sock);
  return (-1);

#else
  fd_set readfds;
  int    s, res;
  u_long thetime;
  struct sockaddr_in from;
  int    fromlen;

  s = socket (AF_INET, SOCK_DGRAM, 0);
  if (s < 0)
     return (-1);

  addrp->sin_family = AF_INET;
  addrp->sin_port   = htons (IPPORT_TIMESERVER);

  res = sendto (s, (char *)&thetime, sizeof(thetime), 0,
                (struct sockaddr *)addrp, sizeof(*addrp));
  if (res < 0)
  {
    do_close (s);
    return (-1);
  }
  do
  {
    FD_ZERO (&readfds);
    FD_SET (s, &readfds);
    res = select (_rpc_dtablesize(),&readfds,NULL,NULL,timeout);
  }
  while (res < 0 && errno == EINTR);

  if (res <= 0)
  {
    if (res == 0)
       errno = ETIMEDOUT;
    do_close (s);
    return (-1);
  }
  fromlen = sizeof (from);
  res = recvfrom (s, (char *)&thetime, sizeof(thetime), 0,
                  (struct sockaddr *)&from, &fromlen);
  do_close (s);
  if (res < 0)
     return (-1);
  
  if (res != sizeof(thetime))
  {
    errno = EIO;
    return (-1);
  }
  thetime = ntohl (thetime);
  timep->tv_sec  = thetime - TOFFSET;
  timep->tv_usec = 0;
  return (0);

#endif
}

