/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */
#if !defined(lint) && defined(SCCSIDS)
static char sccsid[] = "@(#)svc.c 1.41 87/10/13 Copyr 1984 Sun Micro";
#endif

/*
 * svc.c, Server-side remote procedure call interface.
 *
 * There are two sets of procedures here.  The xprt routines are
 * for handling transport handles.  The svc routines handle the
 * list of service routines.
 *
 * Copyright (C) 1984, Sun Microsystems, Inc.
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <rpc/rpc.h>
#include <rpc/pmap_cln.h>
#include <sys/socket.h>

#include "rpc_mac.h"

static SVCXPRT **xports;

#define NULL_SVC     NULL
#define RQCRED_SIZE  400    /* this size is excessive */

/*
 * The services list
 * Each entry represents a set of procedures (an rpc program).
 * The dispatch routine takes request structs and runs the
 * apropriate procedure.
 */
static struct svc_callout {
       struct svc_callout *sc_next;
       u_long              sc_prog;
       u_long              sc_vers;
       void              (*sc_dispatch)();
     } *svc_head;

static struct svc_callout *svc_find (u_long, u_long, struct svc_callout**);

/****************  SVCXPRT related stuff *****************/

/*
 * Activate a transport handle.
 */
void xprt_register (SVCXPRT *xprt)
{
  int sock = xprt->xp_sock;

  if (xports == NULL)
      xports = malloc (FD_SETSIZE * sizeof(SVCXPRT*));

  if (sock < _rpc_dtablesize())
  {
    xports[sock] = xprt;
    FD_SET (sock, &svc_fdset);
  }
}

/*
 * De-activate a transport handle. 
 */
void xprt_unregister (SVCXPRT *xprt)
{ 
  int sock = xprt->xp_sock;

  if ((sock < _rpc_dtablesize()) && (xports[sock] == xprt))
  {
    xports[sock] = NULL;
    FD_CLR (sock, &svc_fdset);
  }
}


/*********************** CALLOUT list related stuff **************/

/*
 * Add a service program to the callout list.
 * The dispatch routine will be called when a rpc request for this
 * program number comes in.
 */
bool_t svc_register (xprt, prog, vers, dispatch, protocol)
  SVCXPRT *xprt;
  u_long   prog;
  u_long   vers;
  void   (*dispatch)();
  int      protocol;
{
  struct svc_callout *prev;
  struct svc_callout *s;

  if ((s = svc_find(prog, vers, &prev)) != NULL_SVC)
  {
    if (s->sc_dispatch == dispatch)
       goto pmap_it;  /* he is registering another xptr */
    return (FALSE);
  }
  s = malloc (sizeof(struct svc_callout));
  if (!s)
    return (FALSE);

  s->sc_prog     = prog;
  s->sc_vers     = vers;
  s->sc_dispatch = dispatch;
  s->sc_next     = svc_head;
  svc_head = s;

pmap_it:
  /* now the information with the local binder service */
  if (protocol)
     return pmap_set (prog, vers, protocol, xprt->xp_port);

  return (TRUE);
}

/*
 * Remove a service program from the callout list.
 */
void svc_unregister (u_long prog, u_long vers)
{
  struct svc_callout *prev;
  struct svc_callout *s;

  if ((s = svc_find(prog,vers,&prev)) == NULL_SVC)
     return;
  if (prev == NULL_SVC)
       svc_head = s->sc_next;
  else prev->sc_next = s->sc_next;

  s->sc_next = NULL_SVC;
  free (s);
  /* now unthe information with the local binder service */
  pmap_unset (prog, vers);
}

/*
 * Search the callout list for a program number, return the callout
 * struct.
 */
static struct svc_callout *svc_find (u_long prog, u_long vers,
                                     struct svc_callout **prev)
{
  struct svc_callout *s;
  struct svc_callout *p = NULL_SVC;

  for (s = svc_head; s != NULL_SVC; s = s->sc_next)
  {
    if (s->sc_prog == prog && s->sc_vers == vers)
       goto done;
    p = s;
  }
done:
  *prev = p;
  return (s);
}

/******************** REPLY GENERATION ROUTINES  *************/

/*
 * Send a reply to an rpc request
 */
bool_t svc_sendreply (SVCXPRT *xprt, xdrproc_t xdr_results, char *xdr_location)
{
  struct rpc_msg rply; 

  rply.rm_direction                 = REPLY;
  rply.rm_reply.rp_stat             = MSG_ACCEPTED;
  rply.acpted_rply.ar_verf          = xprt->xp_verf;
  rply.acpted_rply.ar_stat          = SUCCESS;
  rply.acpted_rply.ar_results.where = (caddr_t) xdr_location;
  rply.acpted_rply.ar_results.proc  = xdr_results;
  return SVC_REPLY (xprt, &rply);
}

/*
 * No procedure error reply
 */
void svcerr_noproc (SVCXPRT *xprt)
{
  struct rpc_msg rply;

  rply.rm_direction        = REPLY;
  rply.rm_reply.rp_stat    = MSG_ACCEPTED;
  rply.acpted_rply.ar_verf = xprt->xp_verf;
  rply.acpted_rply.ar_stat = PROC_UNAVAIL;
  SVC_REPLY (xprt, &rply);
}

/*
 * Can't decode args error reply
 */
void svcerr_decode (SVCXPRT *xprt)
{
  struct rpc_msg rply; 

  rply.rm_direction        = REPLY;
  rply.rm_reply.rp_stat    = MSG_ACCEPTED;
  rply.acpted_rply.ar_verf = xprt->xp_verf;
  rply.acpted_rply.ar_stat = GARBAGE_ARGS;
  SVC_REPLY (xprt, &rply);
}

/*
 * Some system error
 */
void svcerr_systemerr (SVCXPRT *xprt)
{
  struct rpc_msg rply; 

  rply.rm_direction        = REPLY;
  rply.rm_reply.rp_stat    = MSG_ACCEPTED;
  rply.acpted_rply.ar_verf = xprt->xp_verf;
  rply.acpted_rply.ar_stat = SYSTEM_ERR;
  SVC_REPLY (xprt, &rply);
}

/*
 * Authentication error reply
 */
void svcerr_auth (SVCXPRT *xprt, enum auth_stat why)
{
  struct rpc_msg rply;

  rply.rm_direction        = REPLY;
  rply.rm_reply.rp_stat    = MSG_DENIED;
  rply.rjcted_rply.rj_stat = AUTH_ERROR;
  rply.rjcted_rply.rj_why  = why;
  SVC_REPLY (xprt, &rply);
}

/*
 * Auth too weak error reply
 */
void svcerr_weakauth (SVCXPRT *xprt)
{
  svcerr_auth (xprt, AUTH_TOOWEAK);
}

/*
 * Program unavailable error reply
 */
void svcerr_noprog (SVCXPRT *xprt)
{
  struct rpc_msg rply;  

  rply.rm_direction        = REPLY;
  rply.rm_reply.rp_stat    = MSG_ACCEPTED;
  rply.acpted_rply.ar_verf = xprt->xp_verf;  
  rply.acpted_rply.ar_stat = PROG_UNAVAIL;
  SVC_REPLY (xprt, &rply);
}

/*
 * Program version mismatch error reply
 */
void svcerr_progvers (SVCXPRT *xprt, u_long low_vers, u_long high_vers)
{
  struct rpc_msg rply;

  rply.rm_direction             = REPLY;
  rply.rm_reply.rp_stat         = MSG_ACCEPTED;
  rply.acpted_rply.ar_verf      = xprt->xp_verf;
  rply.acpted_rply.ar_stat      = PROG_MISMATCH;
  rply.acpted_rply.ar_vers.low  = low_vers;
  rply.acpted_rply.ar_vers.high = high_vers;
  SVC_REPLY (xprt, &rply);
}

/******************** SERVER INPUT STUFF **********************/

/*
 * Get server side input from some transport.
 *
 * Statement of authentication parameters management:
 * This function owns and manages all authentication parameters, specifically
 * the "raw" parameters (msg.rm_call.cb_cred and msg.rm_call.cb_verf) and
 * the "cooked" credentials (rqst->rq_clntcred).
 * However, this function does not know the structure of the cooked
 * credentials, so it make the following assumptions: 
 *   a) the structure is contiguous (no pointers), and
 *   b) the cred structure size does not exceed RQCRED_SIZE bytes. 
 * In all events, all three parameters are freed upon exit from this routine.
 * The storage is trivially management on the call stack in user land, but
 * is mallocated in kernel land.
 */

void svc_getreq (int rdfds)
{
  fd_set readfds;

  FD_ZERO (&readfds);
  readfds.fd_bits[0] = rdfds;
  svc_getreqset (&readfds);
}

void svc_getreqset (fd_set *readfds)
{
  enum    xprt_stat stat;
  struct  rpc_msg   msg;
  struct  svc_req   r;
  int     prog_found, bit, setsize, sock;
  u_long  low_vers, high_vers, mask, *maskp;
  SVCXPRT *xprt;
  char    cred_area [2*MAX_AUTH_BYTES + RQCRED_SIZE];

  msg.rm_call.cb_cred.oa_base = (caddr_t) cred_area;
  msg.rm_call.cb_verf.oa_base = (caddr_t) (cred_area + MAX_AUTH_BYTES);
  r.rq_clntcred               = (caddr_t) (cred_area + 2*MAX_AUTH_BYTES);

  setsize = _rpc_dtablesize();
  maskp   = (u_long*) readfds->fd_bits;

  for (sock = 0; sock < setsize; sock += (FD_SETSIZE+7) / 8)
  {
    for (mask = *maskp++; (bit = ffs(mask)) != 0; mask ^= (1 << (bit - 1)))
    {
      /* sock has input waiting */
      xprt = xports [sock + bit - 1];

      /* now receive msgs from xprtprt (support batch calls) */
      do
      {
        if (SVC_RECV(xprt,&msg))
        {
          /* now find the exported program and call it */
          struct svc_callout *s;

#if 0  // !! no authentication
          enum   auth_stat why;

          r.rq_xprt = xprt;
          r.rq_prog = msg.rm_call.cb_prog;
          r.rq_vers = msg.rm_call.cb_vers;
          r.rq_proc = msg.rm_call.cb_proc;
          r.rq_cred = msg.rm_call.cb_cred;

          /* first authenticate the message */
          if ((why = _authenticate(&r, &msg)) != AUTH_OK)
          {
            svcerr_auth (xprt, why);
            goto call_done;
          }
#endif
          /* now match message with a registered service*/
          prog_found = FALSE;
          low_vers   = (u_long)-1;
          high_vers  = 0;

          for (s = svc_head; s != NULL_SVC; s = s->sc_next)
          {
            if (s->sc_prog == r.rq_prog)
            {
              if (s->sc_vers == r.rq_vers)
              {
                (*s->sc_dispatch) (&r,xprt);
                goto call_done;
              }
              /* found correct version */
              prog_found = TRUE;
              if (s->sc_vers < low_vers)
                 low_vers = s->sc_vers;
              if (s->sc_vers > high_vers)
                 high_vers = s->sc_vers;
            }
          }
          /*
           * if we got here, the program or version
           * is not served ...
           */
          if (prog_found)
               svcerr_progvers (xprt, low_vers, high_vers);
          else svcerr_noprog (xprt);
        }

      call_done:
        if ((stat = SVC_STAT(xprt)) == XPRT_DIED)
        {
          SVC_DESTROY (xprt);
          break;
        }
      }
      while (stat == XPRT_MOREREQS);
    }
  }
}
